import os
import numpy as np
import pandas as pd
from subprocess import check_call
from typing import List, Tuple, Dict, Set, Union, Optional

from pegasusio import timer
from pegasusio import MultimodalData, AggrData
from pegasusio import infer_file_type, read_input

import logging
logger = logging.getLogger(__name__)


def _find_digits(value: str) -> Tuple[str, int]:
    pos = len(value) - 1
    while pos >= 0 and value[pos].isdigit():
        pos -= 1
    pos += 1
    assert pos < len(value)
    return (value[:pos], int(value[pos:]))


def _parse_restriction_string(rstr: str) -> Tuple[str, bool, Set[str]]:
    pos = rstr.index(":")
    name = rstr[:pos]
    isin = True
    if rstr[pos + 1] == "~":
        isin = False
        pos += 1
    content = set()
    for item in rstr[pos + 1:].split(","):
        values = item.split("-")
        if len(values) == 1:
            content.add(values[0])
        else:
            prefix, fr = _find_digits(values[0])
            assert values[1].isdigit()
            to = int(values[1]) + 1
            for i in range(fr, to):
                content.add(prefix + str(i))
    return (name, isin, content)


def _parse_genome_string(genome_str: str) -> Tuple[str, Dict[str, str]]:
    genome = None
    genome_dict = {}

    if genome_str is not None:
        fields = genome_str.split(",")
        for field in fields:
            items = field.split(":")
            if len(items) == 1:
                genome = items[0]
            else:
                genome_dict[items[0]] = items[1]

    return genome, genome_dict


@timer(logger=logger)
def aggregate_matrices(
    csv_file: Union[str, Dict[str, np.ndarray], pd.DataFrame],
    restrictions: Optional[Union[List[str], str]] = [],
    attributes: Optional[Union[List[str], str]] = [],
    default_ref: Optional[str] = None,
    append_sample_name: Optional[bool] = True,
    select_singlets: Optional[bool] = False,
    remap_string: Optional[str] = None,
    subset_string: Optional[str] = None,
    min_genes: Optional[int] = None,
    max_genes: Optional[int] = None,
    min_umis: Optional[int] = None,
    max_umis: Optional[int] = None,
    mito_prefix: Optional[str] = None,
    percent_mito: Optional[float] = None,
) -> MultimodalData:
    """Aggregate channel-specific count matrices into one big count matrix.

    This function takes as input a csv_file, which contains at least 2 columns — Sample, sample name; Location, file that contains the count matrices (e.g. filtered_gene_bc_matrices_h5.h5), and merges matrices from the same genome together. If multi-modality exists, a third Modality column might be required. An aggregated Multimodal Data will be returned.

    If csv_file is a dictionary, it can contains an alternative 2 columns - Sample, sample name; Object, Multimodal data object. In this case, the objects will be merged into one data object. In addition, aggregate_matrices will make copies instead of editing the objects.

    The csv_file can optionally contain two columns - nUMI and nGene. These two columns define minimum number of UMIs and genes for cell selection for each sample. The values in these two columns overwrite the min_genes and min_umis arguments.

    Parameters
    ----------

    csv_file : `str`
        The CSV file containing information about each channel. Alternatively, a dictionary or pd.Dataframe can be passed.
    restrictions : `list[str]` or `str`, optional (default: [])
        A list of restrictions used to select channels, each restriction takes the format of name:value,…,value or name:~value,..,value, where ~ refers to not. If only one restriction is provided, it can be provided as a string instead of a list.
    attributes : `list[str]` or `str`, optional (default: [])
        A list of attributes need to be incorporated into the output count matrix. If only one attribute is provided, this attribute can be provided as a string instead of a list.
    default_ref : `str`, optional (default: None)
        Default reference name to use. If there is no Reference column in the csv_file, a Reference column will be added with default_ref as its value. This argument can also be used for replacing genome names. For example, if default_ref is 'hg19:GRCh38,GRCh38', we will change any genome with name 'hg19' to 'GRCh38' and if no genome is provided, 'GRCh38' is the default.
    append_sample_name : `bool`, optional (default: True)
        By default, append sample_name to each channel. Turn this option off if each channel has distinct barcodes.
    select_singlets : `bool`, optional (default: False)
        If we have demultiplexed data, turning on this option will make pegasus only include barcodes that are predicted as singlets.
    remap_string: ``str``, optional, default ``None``
        Remap singlet names using <remap_string>, where <remap_string> takes the format "new_name_i:old_name_1,old_name_2;new_name_ii:old_name_3;...". For example, if we hashed 5 libraries from 3 samples sample1_lib1, sample1_lib2, sample2_lib1, sample2_lib2 and sample3, we can remap them to 3 samples using this string: "sample1:sample1_lib1,sample1_lib2;sample2:sample2_lib1,sample2_lib2". In this way, the new singlet names will be in metadata field with key 'assignment', while the old names will be kept in metadata field with key 'assignment.orig'.
    subset_string: ``str``, optional, default ``None``
        If select singlets, only select singlets in the <subset_string>, which takes the format "name1,name2,...". Note that if --remap-singlets is specified, subsetting happens after remapping. For example, we can only select singlets from sampe 1 and 3 using "sample1,sample3".
    min_genes: ``int``, optional, default: None
       Only keep cells with at least ``min_genes`` genes.
    max_genes: ``int``, optional, default: None
       Only keep cells with less than ``max_genes`` genes.
    min_umis: ``int``, optional, default: None
       Only keep cells with at least ``min_umis`` UMIs.
    max_umis: ``int``, optional, default: None
       Only keep cells with less than ``max_umis`` UMIs.
    mito_prefix: ``str``, optional, default: None
       Prefix for mitochondrial genes.
    percent_mito: ``float``, optional, default: None
       Only keep cells with percent mitochondrial genes less than ``percent_mito`` % of total counts. Only when both mito_prefix and percent_mito set, the mitochondrial filter will be triggered.

    Returns
    -------
    `MultimodalData` object.
        The aggregated count matrix as an MultimodalData object.

    Examples
    --------
    >>> data = aggregate_matrix('example.csv', restrictions=['Source:pbmc', 'Donor:1'], attributes=['Source', 'Platform', 'Donor'])
    """
    if isinstance(csv_file, str):
        df = pd.read_csv(csv_file, header=0, index_col=False) # load sample sheet
    elif isinstance(csv_file, dict):
        df = pd.DataFrame(csv_file)
    else:
        df = csv_file

    # Remove duplicated items
    if isinstance(restrictions, str):
        restrictions = [restrictions]
    restrictions = set(restrictions)
    if isinstance(attributes, str):
        attributes = [attributes]
    attributes = set(attributes)

    # Select data
    rvec = [_parse_restriction_string(x) for x in restrictions]

    idx = pd.Series([True] * df.shape[0], index=df.index, name="selected")
    for name, isin, content in rvec:
        assert name in df.columns
        if isin:
            idx = idx & df[name].isin(content)
        else:
            idx = idx & (~(df[name].isin(content)))

    if idx.sum() == 0:
        raise ValueError("No data pass the restrictions!")

    df = df.loc[idx].sort_values(by = "Sample") # sort by sample_name

    # parse default_ref
    def_genome, genome_dict = _parse_genome_string(default_ref)

    # Load data
    tot = 0
    dest_paths = [] # record localized file paths so that we can remove them later
    curr_sample = ""
    curr_row = curr_data = None
    aggrData = AggrData()

    for idx_num, row in df.iterrows():
        if "Object" in row:
            data = row["Object"].copy()
        else:
            assert "Location" in row, f"Row of sample '{row['Sample']}' must contain a 'Location' column!"
            assert not row.isnull().values.any(), f"Row of sample '{row['Sample']}' has one or more NaN/NA values!"

            input_file = os.path.expanduser(os.path.expandvars(row["Location"].rstrip(os.sep))) # extend all user variables
            file_type, copy_path, copy_type = infer_file_type(input_file) # infer file type

            if row["Location"].lower().startswith('gs://'): # if Google bucket
                base_name = os.path.basename(copy_path)
                dest_path = f"{idx_num}_tmp_{base_name}" # id_num will make sure dest_path is unique in the sample sheet
                if not os.path.exists(dest_path):  # if dest_path exists, we may try to localize it once and may have the file cached
                    if copy_type == "directory":
                        check_call(["mkdir", "-p", dest_path])
                        call_args = ["gsutil", "-m", "rsync", "-r", copy_path, dest_path]
                    else:
                        call_args = ["gsutil", "-m", "cp", copy_path, dest_path]
                    check_call(call_args)
                dest_paths.append(dest_path)

                if input_file == copy_path:
                    input_file = dest_path
                else:
                    input_file = os.path.join(dest_path, os.path.basename(input_file))

            genome = row.get("Reference", None)
            if (genome is not None) and (not isinstance(genome, str)): # to avoid NaN
                genome = None
            if genome is None:
                genome = def_genome
            modality = row.get("Modality", None)
            data = read_input(input_file, file_type = file_type, genome = genome, modality = modality)

        if len(genome_dict) > 0:
            data._update_genome(genome_dict)

        if row["Sample"] != curr_sample:
            if curr_data is not None:
                curr_data._propogate_genome()
                curr_data.filter_data(select_singlets = select_singlets, remap_string = remap_string, subset_string = subset_string, min_genes = min_genes, max_genes = max_genes, min_umis = min_umis, max_umis = max_umis, mito_prefix = mito_prefix, percent_mito = percent_mito)
                curr_data._update_barcode_metadata_info(curr_row, attributes, append_sample_name)
                aggrData.add_data(curr_data)
            curr_data = data
            curr_row = row
            curr_sample = row["Sample"]
            if "nUMI" in row:
                min_umis = row["nUMI"]
            if "nGene" in row:
                min_genes = row["nGene"]
        else:
            curr_data.update(data)

        tot += 1

    if curr_data is not None:
        curr_data._propogate_genome()
        curr_data.filter_data(select_singlets = select_singlets, remap_string = remap_string, subset_string = subset_string, min_genes = min_genes, max_genes = max_genes, min_umis = min_umis, max_umis = max_umis, mito_prefix = mito_prefix, percent_mito = percent_mito)
        curr_data._update_barcode_metadata_info(curr_row, attributes, append_sample_name)
        aggrData.add_data(curr_data)

    # Merge data
    aggregated_data = aggrData.aggregate()
    attributes.add("Channel")
    aggregated_data._convert_attributes_to_categorical(attributes)
    logger.info(f"Aggregated {tot} files.")

    # Delete temporary file
    if len(dest_paths) > 0:
        for dest_path in dest_paths:
            check_call(["rm", "-rf", dest_path])
        logger.info("Temporary files are deleted.")

    return aggregated_data



from .Base import Base
from pegasusio import aggregate_matrices, write_output


class AggregateMatrix(Base):
    """
Aggregate multiple single-modality or multi-modality data into one big MultimodalData object and write it back to disk as a zipped Zarr file.

Usage:
  pegasusio aggregate_matrix <csv_file> <output_name> [--restriction <restriction>... options]
  pegasusio aggregate_matrix -h

Arguments:
  csv_file           csv_file should contains at least 2 columns — Sample, sample name; Location, file that contains the count matrices (e.g. filtered_gene_bc_matrices_h5.h5), and merges matrices from the same genome together. If multi-modality exists, a third Modality column might be required. The csv_file can optionally contain two columns - nUMI and nGene. These two columns define minimum number of UMIs and genes for cell selection for each sample. The values in these two columns overwrite the min_genes and min_umis arguments.
  output_name        The output file name.

Options:
  --restriction <restriction>...           Select data that satisfy all restrictions. Each restriction takes the format of name:value,...,value or name:~value,..,value, where ~ refers to not. You can specifiy multiple restrictions by setting this option multiple times.
  --attributes <attributes>                Specify a comma-separated list of outputted attributes. These attributes should be column names in the csv file.
  --default-reference <reference>          If sample count matrix is in either DGE, mtx, csv, tsv or loom format and there is no Reference column in the csv_file, use <reference> as the reference. This option can also be used for replacing genome names. For example, if <reference> is 'hg19:GRCh38,GRCh38', we will change any genome with name 'hg19' to 'GRCh38' and if no genome is provided, 'GRCh38' is the default.
  --select-only-singlets                   If we have demultiplexed data, turning on this option will make pegasusio only include barcodes that are predicted as singlets.
  --remap-singlets <remap_string>          Remap singlet names using <remap_string>, where <remap_string> takes the format "new_name_i:old_name_1,old_name_2;new_name_ii:old_name_3;...". For example, if we hashed 5 libraries from 3 samples sample1_lib1, sample1_lib2, sample2_lib1, sample2_lib2 and sample3, we can remap them to 3 samples using this string: "sample1:sample1_lib1,sample1_lib2;sample2:sample2_lib1,sample2_lib2". In this way, the new singlet names will be in metadata field with key 'assignment', while the old names will be kept in metadata field with key 'assignment.orig'.
  --subset-singlets <subset_string>        If select singlets, only select singlets in the <subset_string>, which takes the format "name1,name2,...". Note that if --remap-singlets is specified, subsetting happens after remapping. For example, we can only select singlets from sampe 1 and 3 using "sample1,sample3".
  --min-genes <number>                     Only keep cells with at least <number> of genes.
  --max-genes <number>                     Only keep cells with less than <number> of genes. 
  --min-umis <number>                      Only keep cells with at least <number> of UMIs.
  --max-umis <number>                      Only keep cells with less than <number> of UMIs.
  --mito-prefix <prefix>                   Prefix for mitochondrial genes. If multiple prefixes are provided, separate them by comma (e.g. "MT-,mt-").
  --percent-mito <percent>                 Only keep cells with mitochondrial percent less than <percent>%. Only when both mito_prefix and percent_mito set, the mitochondrial filter will be triggered.
  --no-append-sample-name                  Turn this option on if you do not want to append sample name in front of each sample's barcode (concatenated using '-').
  -h, --help                               Print out help information.

Outputs:
  output_name.zarr.zip        A zipped Zarr file containing aggregated data.

Examples:
  pegasusio aggregate_matrix --restriction Source:BM,CB --restriction Individual:1-8 --attributes Source,Platform count_matrix.csv aggr_data
    """

    def execute(self):
        data = aggregate_matrices(
            self.args["<csv_file>"],
            restrictions=self.args["--restriction"],
            attributes=self.split_string(self.args["--attributes"]),
            default_ref=self.args["--default-reference"],
            append_sample_name=not self.args["--no-append-sample-name"],
            select_singlets=self.args["--select-only-singlets"],
            remap_string=self.args["--remap-singlets"],
            subset_string=self.args["--subset-singlets"],
            min_genes=self.convert_to_int(self.args["--min-genes"]),
            max_genes=self.convert_to_int(self.args["--max-genes"]),
            min_umis=self.convert_to_int(self.args["--min-umis"]),
            max_umis=self.convert_to_int(self.args["--max-umis"]),
            mito_prefix=self.args["--mito-prefix"],
            percent_mito=self.convert_to_float(self.args["--percent-mito"])
        )
        write_output(data, self.args["<output_name>"] + ".zarr.zip")



import sys
import logging
import warnings


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)

warnings.filterwarnings("ignore", category=FutureWarning, module='anndata')


modalities = ['rna', 'citeseq', 'hashing', 'tcr', 'bcr', 'crispr', 'atac', 'cyto', 'nanostring']

from .decorators import timer, run_gc

from .unimodal_data import UnimodalData
from .vdj_data import VDJData
from .citeseq_data import CITESeqData
from .cyto_data import CytoData
from .nanostring_data import NanostringData
from .qc_utils import calc_qc_filters, apply_qc_filters, DictWithDefault
from .multimodal_data import MultimodalData
from .aggr_data import AggrData, _fillna
from .readwrite import infer_file_type, read_input, write_output, write_scp_file
from .data_aggregation import aggregate_matrices

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:  # < Python 3.8: Use backport module
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version(__name__)
    del version
except PackageNotFoundError:
    pass



import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from typing import List, Dict, Union

import logging
logger = logging.getLogger(__name__)

import anndata
from pegasusio import UnimodalData
from .views import INDEX, _parse_index, UnimodalDataView


class NanostringData(UnimodalData):
    _matrix_keywords = ["RawData", "Q3Norm", "HKNorm", "LogMatrix"] # all in dense format, np.ndarray

    def __init__(
        self,
        barcode_metadata: Union[dict, pd.DataFrame],
        feature_metadata: Union[dict, pd.DataFrame],
        matrices: Dict[str, np.ndarray],
        metadata: dict,
        barcode_multiarrays: Dict[str, np.ndarray] = dict(),
        feature_multiarrays: Dict[str, np.ndarray] = dict(),
        barcode_multigraphs: Dict[str, csr_matrix] = dict(),
        feature_multigraphs: Dict[str, csr_matrix] = dict(),
        cur_matrix: str = "Q3Norm",
    ) -> None:
        assert metadata["modality"] == "nanostring"
        super().__init__(barcode_metadata, feature_metadata, matrices, metadata, barcode_multiarrays, feature_multiarrays, barcode_multigraphs, feature_multigraphs, cur_matrix)


    def from_anndata(self, data: anndata.AnnData, genome: str = None, modality: str = None) -> None:
        raise ValueError("Cannot convert an AnnData object to a NanostringData object!")


    def to_anndata(self) -> anndata.AnnData:
        raise ValueError("Cannot convert a NanostringData object ot an AnnData object!")


    def __getitem__(self, index: INDEX) -> UnimodalDataView:
        barcode_index, feature_index = _parse_index(self, index)
        return UnimodalDataView(self, barcode_index, feature_index, self._cur_matrix, obj_name = "NanostringData")


    def norm_hk(self, select: bool = True) -> None:
        """ Normalize raw protein data using house keeping (HK) genes.
            See https://genomebiology.biomedcentral.com/articles/10.1186/gb-2002-3-7-research0034 for geometric mean normalization.

        Parameters
        ----------
        select: ``bool``, optional, default: ``True``
            If True, select the normalized matrix as the major matrix (X).

        Returns
        -------
        ``None``

        Update ``self.matrices``:
            * ``self.matrices['HKNorm']``: np.ndarray.

        Examples
        --------
        >>> nanostring_data.norm_hk()
        """
        from scipy.stats.mstats import gmean

        gms = gmean(self.obsm["controls"], axis = 1)
        norm_factors = gmean(gms) / gms
        self.matrices["HKNorm"] = self.matrices["RawData"] * norm_factors.reshape(-1, 1)
        self.obs["norm_factor"] = norm_factors

        if select:
            self._cur_matrix = "HKNorm"


    def log_transform(self, select: bool = True) -> None:
        """Conduct log transformation on the selected matrix: log(x + 1). Selected matrix can be either Q3Norm or HKNorm

        Add log-transformed matrix 'LogMatrix'.

        Parameters
        ----------
        select: ``bool``, optional, default: ``True``
            If True, select the transformed matrix as the major matrix (X).

        Returns
        -------
        ``None``

        Update ``self.matrices``:
            * ``self.matrices['LogMatrix']``: np.ndarray.

        Examples
        --------
        >>> nanostring_data.log_norm()
        """
        if self.current_matrix() not in ["Q3Norm", "HKNorm"]:
            raise ValueError("Either Q3Norm or HKNorm matrix must be selected in order to run the 'log_norm' function!")

        self.matrices["LogMatrix"] = np.log1p(self.X)

        if select:
            self._cur_matrix = "LogMatrix"



import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, vstack, coo_matrix
from typing import List, Dict, Union
from collections import defaultdict
from pandas.api.types import is_categorical_dtype, is_string_dtype
from natsort import natsorted

import logging
logger = logging.getLogger(__name__)

from pegasusio import timer, run_gc
from pegasusio import UnimodalData, CITESeqData, CytoData, VDJData, MultimodalData


def _fillna(df: pd.DataFrame) -> None:
    """ Fill NA with default values """
    isna = df.isna()
    hasna = isna.sum(axis=0)
    for column in hasna.index[hasna > 0]:
        if is_categorical_dtype(df[column]):
            df[column] = df[column].astype(str)

        default_value = None
        if df[column].dtype.kind == "O" and type(df[column].unique()[0]) == bool:
            default_value = False
        elif df[column].dtype.kind in {"i", "u", "f", "c"}:
            default_value = 0
        elif df[column].dtype.kind == "S":
            default_value = b""
        elif df[column].dtype.kind in {"O", "U"}:
            default_value = ""
        else:
            raise ValueError(f"{column} has unsupported dtype {df[column].dtype}!")

        df.loc[isna[column], column] = default_value

        if type(default_value) == bool:
            df[column] = df[column].astype(bool)

        if df[column].dtype.kind == "f":
            int_type = getattr(np, df[column].dtype.name.replace('float', 'int'))
            values = df[column].values
            int_values = values.astype(int_type)
            if np.array_equal(values, int_values):
                df[column] = int_values


def _check_categorical(df:pd.DataFrame) -> None:
    for col in df.columns:
        if not is_categorical_dtype(df[col]) and is_string_dtype(df[col]):
            keywords = set(df[col])
            if len(keywords) <= df.shape[0] / 10.0: # at least 10x reduction
                df[col] = pd.Categorical(df[col], categories = natsorted(keywords))


class AggrData:
    def __init__(self):
        self.aggr = defaultdict(list)


    def add_data(self, data: MultimodalData) -> None:
        for key in data.list_data():
            self.aggr[key].append(data.get_data(key))


    @run_gc
    def _merge_matrices(self, feature_metadata: pd.DataFrame, unilist: List[UnimodalData], modality: str) -> Dict[str, csr_matrix]:
        """ Merge all matrices together """
        f2idx = pd.Series(data=range(feature_metadata.shape[0]), index=feature_metadata.index)

        mat_keys = set()
        no_reorg = True
        for unidata in unilist:
            mat_keys.update(unidata.matrices)
            if no_reorg and (feature_metadata.shape[0] > unidata.feature_metadata.shape[0] or (feature_metadata.index != unidata.feature_metadata.index).sum() > 0):
                no_reorg = False

        if modality == "bcr" or modality == "tcr":
            for key in VDJData._uns_keywords:
                mat_keys.discard(key[1:])


        matrices = {}
        if no_reorg:
            for mat_key in mat_keys:
                mat_list = []
                for unidata in unilist:
                    mat = unidata.matrices.pop(mat_key, None)
                    if mat is not None:
                        mat_list.append(mat)
                matrices[mat_key] = vstack(mat_list) if modality != "cyto" else np.vstack(mat_list)
        else:
            colmap = []
            for unidata in unilist:
                colmap.append(f2idx[unidata.feature_metadata.index].values)

            if modality == "cyto":
                # matrices are dense.
                for mat_key in mat_keys:
                    mat_list = []
                    for i, unidata in enumerate(unilist):
                        mat = unidata.matrices.pop(mat_key, None)
                        if mat is not None:
                            newmat = np.zeros((mat.shape[0], feature_metadata.shape[0]), dtype = mat.dtype)
                            newmat[:, colmap[i]] = mat
                            mat_list.append(newmat)
                    matrices[mat_key] = np.vstack(mat_list)
            else:
                for mat_key in mat_keys:
                    data_list = []
                    row_list = []
                    col_list = []
                    row_base = 0
                    for i, unidata in enumerate(unilist):
                        mat = unidata.matrices.pop(mat_key, None)
                        if mat is not None:
                            mat = mat.tocoo(copy = False) # convert to coo format
                            data_list.append(mat.data)
                            row_list.append(mat.row + row_base)
                            col_list.append(colmap[i][mat.col])
                            row_base += mat.shape[0]
                    data = np.concatenate(data_list)
                    row = np.concatenate(row_list)
                    col = np.concatenate(col_list)
                    matrices[mat_key] = coo_matrix((data, (row, col)), shape=(row_base, feature_metadata.shape[0])).tocsr(copy = False)

        return matrices


    @run_gc
    def _vdj_update_metadata_matrices(self, metadata: dict, matrices: dict, unilist: List[UnimodalData]) -> None:
        metadata.pop("uns_dict", None)
        for key in VDJData._uns_keywords:
            values = set()
            for unidata in unilist:
                values.update(unidata.metadata[key])
            values.discard("None") # None must be 0 for sparse matrix
            metadata[key] = np.array(["None"] + list(values), dtype = np.object)
            val2num = dict(zip(metadata[key], range(metadata[key].size)))

            mat_key = key[1:]
            mat_list = []
            for unidata in unilist:
                mat = unidata.matrices.pop(mat_key)
                old2new = np.array([val2num[val] for val in unidata.metadata[key]], dtype = np.int32)
                mat.data = old2new[mat.data]
                mat_list.append(mat)
            matrices[mat_key] = vstack(mat_list)


    @run_gc
    def _aggregate_unidata(self, unilist: List[UnimodalData]) -> UnimodalData:
        if len(unilist) == 1:
            del unilist[0].metadata["_sample"]
            return unilist[0]

        modality = unilist[0].get_modality()

        barcode_metadata_dfs = [unidata.barcode_metadata for unidata in unilist]
        barcode_metadata = pd.concat(barcode_metadata_dfs, axis=0, sort=False, copy=False)
        _fillna(barcode_metadata)
        _check_categorical(barcode_metadata)

        var_dict = {}
        for unidata in unilist:
            idx = unidata.feature_metadata.columns.difference(["featureid"])
            if idx.size > 0:
                var_dict[unidata.metadata["_sample"]] = unidata.feature_metadata[idx]
                unidata.feature_metadata.drop(columns = idx, inplace = True)

        feature_metadata = unilist[0].feature_metadata
        for other in unilist[1:]:
            keys = ["featurekey"] + feature_metadata.columns.intersection(other.feature_metadata.columns).values.tolist()
            feature_metadata = feature_metadata.merge(other.feature_metadata, on=keys, how="outer", sort=False, copy=False)  # If sort is True, feature keys will be changed even if all channels share the same feature keys.
        _fillna(feature_metadata)
        _check_categorical(feature_metadata)

        matrices = self._merge_matrices(feature_metadata, unilist, modality)

        uns_dict = {}
        metadata = {"genome": unilist[0].metadata["genome"], "modality": unilist[0].metadata["modality"]}
        for unidata in unilist:
            assert unidata.metadata.pop("genome") == metadata["genome"]
            assert unidata.metadata.pop("modality") == metadata["modality"]
            if modality == "citeseq":
                for key in CITESeqData._uns_keywords:
                    unidata.metadata.pop(key, None)
                unidata.metadata.pop("_obs_keys", None)
            elif modality == "cyto":
                for key in CytoData._uns_keywords:
                    unidata.metadata.pop(key, None)
            sample_name = unidata.metadata.pop("_sample")
            if len(unidata.metadata) > 0:
                uns_dict[sample_name] = unidata.metadata.mapping

        if len(var_dict) > 0:
            metadata["var_dict"] = var_dict
        if len(uns_dict) > 0:
            metadata["uns_dict"] = uns_dict


        unidata = None
        if isinstance(unilist[0], CITESeqData):
            unidata = CITESeqData(barcode_metadata, feature_metadata, matrices, metadata)
        elif isinstance(unilist[0], CytoData):
            unidata = CytoData(barcode_metadata, feature_metadata, matrices, metadata)
        elif isinstance(unilist[0], VDJData):
            self._vdj_update_metadata_matrices(metadata, matrices, unilist)
            unidata = VDJData(barcode_metadata, feature_metadata, matrices, metadata)
        else:
            unidata = UnimodalData(barcode_metadata, feature_metadata, matrices, metadata)

        return unidata


    @timer(logger=logger)
    def aggregate(self) -> MultimodalData:
        """ Aggregate all data together """
        data = MultimodalData()
        for key in list(self.aggr):
            unidata = self._aggregate_unidata(self.aggr.pop(key))
            data.add_data(unidata)
        return data



import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, hstack
from typing import List, Dict, Union

import logging
logger = logging.getLogger(__name__)

import anndata
from pegasusio import UnimodalData
from .views import INDEX, _parse_index, UnimodalDataView


# According to communications with Biolegend, isotype control as background is not recommended.

class CITESeqData(UnimodalData):
    _matrix_keywords = ["arcsinh.transformed", "raw.count", "arcsinh.jitter"] # 'arcsinh.jitter' is in dense format, np.ndarray
    _uns_keywords = ["_other_names", "_other_counts"] # '_other' are antibodies that set aside

    def __init__(
        self,
        barcode_metadata: Union[dict, pd.DataFrame],
        feature_metadata: Union[dict, pd.DataFrame],
        matrices: Dict[str, csr_matrix],
        metadata: dict,
        barcode_multiarrays: Dict[str, np.ndarray] = dict(),
        feature_multiarrays: Dict[str, np.ndarray] = dict(),
        barcode_multigraphs: Dict[str, csr_matrix] = dict(),
        feature_multigraphs: Dict[str, csr_matrix] = dict(),
        cur_matrix: str = "raw.count",
    ) -> None:
        assert metadata["modality"] == "citeseq"
        super().__init__(barcode_metadata, feature_metadata, matrices, metadata, barcode_multiarrays, feature_multiarrays, barcode_multigraphs, feature_multigraphs, cur_matrix)


    def from_anndata(self, data: anndata.AnnData, genome: str = None, modality: str = None) -> None:
        raise ValueError("Cannot convert an AnnData object to a CITESeqData object!")


    def to_anndata(self) -> anndata.AnnData:
        raise ValueError("Cannot convert a CITESeqData object ot an AnnData object!")


    def __getitem__(self, index: INDEX) -> UnimodalDataView:
        barcode_index, feature_index = _parse_index(self, index)
        return UnimodalDataView(self, barcode_index, feature_index, self._cur_matrix, obj_name = "CITESeqData")


    def set_aside(self, params: List[str]) -> None:
        """ Move antibodies in params from the raw.count matrix
        """
        assert len(self.matrices) == 1 and "raw.count" in self.matrices
        assert "_other_names" not in self.metadata

        locs = self.feature_metadata.index.get_indexer(params)
        if (locs < 0).sum() > 0:
            raise ValueError(f"Detected unknown antibodies {params[locs < 0]}!")
        self.metadata["_other_names"] = self.feature_metadata.index.values[locs] # with loc: List[int], this should be a copy not a reference
        self.metadata["_other_counts"] = self.matrices["raw.count"][:, locs]

        obs_keys = self.metadata.get("_obs_keys", [])
        obs_keys.append("_other_counts")
        self.metadata["_obs_keys"] = obs_keys

        idx = np.ones(self._shape[1], dtype = bool)
        idx[locs] = False
        self._inplace_subset_var(idx)


    def arcsinh_transform(self, cofactor: float = 5.0, jitter = False, random_state = 0, select: bool = True) -> None:
        """Conduct arcsinh transform on the raw.count matrix.

        Add arcsinh transformed matrix 'arcsinh.transformed'. If jitter == True, instead add a 'arcsinh.jitter' matrix in dense format, jittering by adding a randomized value in U([-0.5, 0.5)). Mimic Cytobank.

        Parameters
        ----------
        cofactor: ``float``, optional, default: ``5.0``
            Cofactor used in cytobank, arcsinh(x / cofactor).

        jitter: ``bool``, optional, default: ``False``
            Add a 'arcsinh.jitter' matrix in dense format, jittering by adding a randomized value in U([-0.5, 0.5)).

        random_state: ``int``, optional, default: ``0``
            Random seed for generating jitters.

        select: ``bool``, optional, default: ``True``
            If True, select the transformed matrix as the major matrix (X).

        Returns
        -------
        ``None``

        Update ``self.matrices``:
            * ``self.matrices['arcsinh.transformed']``: if jitter == False, csr_matrix.
            * ``self.matrices['arcsinh.jitter']``: if jitter == True, np.ndarray.

        Examples
        --------
        >>> citeseq_data.arcsinh_transform(jitter = True)
        """
        if "raw.count" not in self.matrices:
            raise ValueError("raw.count matrix must exist in order to calculate the arcsinh transformed matrix!")

        signal = self.matrices["raw.count"].copy()

        if jitter:
            np.random.seed(random_state)
            jitters = np.random.uniform(low = -0.5, high = 0.5, size = signal.shape)
            signal = np.add(signal.toarray(), jitters, dtype = np.float32)
            self.matrices["arcsinh.jitter"] = np.arcsinh(signal / cofactor, dtype = np.float32)
        else:
            signal.data = np.arcsinh(signal.data / cofactor, dtype = np.float32)
            self.matrices["arcsinh.transformed"] = signal

        if select:
            self._cur_matrix = "arcsinh.jitter" if jitter else "arcsinh.transformed"



import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from collections.abc import MutableMapping
from copy import deepcopy
from natsort import natsorted
from typing import List, Dict, Union, Set, Tuple

import logging
logger = logging.getLogger(__name__)

import anndata

from pegasusio import run_gc
from pegasusio import modalities
from .views import INDEX, _parse_index, UnimodalDataView
from .datadict import DataDict



def _set_genome(metadata: dict, genome: str):
    if genome is not None:
        metadata["genome"] = genome
    elif "genome" not in metadata:
        metadata["genome"] = "unknown"
    elif isinstance(metadata["genome"], np.ndarray):
        assert metadata["genome"].ndim == 1
        metadata["genome"] = metadata["genome"][0]

def _set_modality(metadata: dict, modality: str):
    if modality is not None:
        metadata["modality"] = modality
    elif "modality" not in metadata:
        metadata["modality"] = "rna"


class UnimodalData:
    def __init__(
        self,
        barcode_metadata: Union[dict, pd.DataFrame, anndata.AnnData] = pd.DataFrame(),
        feature_metadata: Union[dict, pd.DataFrame] = pd.DataFrame(),
        matrices: Dict[str, csr_matrix] = dict(),
        metadata: dict = dict(),
        barcode_multiarrays: Dict[str, np.ndarray] = dict(),
        feature_multiarrays: Dict[str, np.ndarray] = dict(),
        barcode_multigraphs: Dict[str, csr_matrix] = dict(),
        feature_multigraphs: Dict[str, csr_matrix] = dict(),
        cur_matrix: str = "X",
        genome: str = None,
        modality: str = None,
    ) -> None:
        """ Note that metadata, barcode_mutiarrays, feature_multiarrays, barcode_multigraphs, feature_multigraphs can be modified.
        """
        if isinstance(barcode_metadata, anndata.AnnData):
            self.from_anndata(barcode_metadata, genome = genome, modality = modality)
            return None

        _set_genome(metadata, genome)
        _set_modality(metadata, modality)

        self.barcode_metadata = barcode_metadata # barcode metadata
        self.feature_metadata = feature_metadata # feature metadata

        if len(self.barcode_metadata) > 0:
            if isinstance(self.barcode_metadata, MutableMapping):
                if "barcodekey" not in self.barcode_metadata:
                    raise ValueError("Cannot locate barcode index barcodekey!")
                barcodekey = self.barcode_metadata.pop("barcodekey")
                self.barcode_metadata = pd.DataFrame(data = self.barcode_metadata, index = pd.Index(barcodekey, name = "barcodekey"))
            else:
                if not isinstance(self.barcode_metadata, pd.DataFrame):
                    raise ValueError(f"Unknown barcode_metadata type: {type(self.barcode_metadata)}!")
                if "barcodekey" in self.barcode_metadata:
                    self.barcode_metadata.set_index("barcodekey", inplace=True)
                elif self.barcode_metadata.index.name != "barcodekey":
                    raise ValueError("Cannot locate barcode index barcodekey!")

        if len(self.feature_metadata) > 0:
            if isinstance(self.feature_metadata, MutableMapping):
                self.feature_metadata = pd.DataFrame(self.feature_metadata)
            if not isinstance(self.feature_metadata, pd.DataFrame):
                raise ValueError(f"Unknown feature_metadata type: {type(self.feature_metadata)}!")
            if "featurekey" in self.feature_metadata:
                self._polish_featurekey(self.feature_metadata, metadata["genome"])
                self.feature_metadata.set_index("featurekey", inplace=True)
            elif self.feature_metadata.index.name != "featurekey":
                raise ValueError("Cannot locate feature index featurekey!")

        self._update_shape()


        self.matrices = DataDict(matrices) # a dictionary of scipy csr matrix

        for key, mat in self.matrices.items():
            if mat.shape[0] != self._shape[0]:
                raise ValueError(f"Wrong number of barcodes : matrix '{key}' has {mat.shape[0]} barcodes, barcodes file has {self._shape[0]} barcodes.")
            if mat.shape[1] != self._shape[1]:
                raise ValueError(f"Wrong number of features : matrix '{key}' has {mat.shape[1]} features, features file has {self._shape[1]} features.")

        if cur_matrix not in matrices.keys():
            raise ValueError("Cannot find the default count matrix to bind to. Please set 'cur_matrix' argument in UnimodalData constructor!")
        self._cur_matrix = cur_matrix # cur_matrix


        # For backword compatibility, check metadata and move arrays and graphs to multiarrays/multigraphs
        for key in list(metadata.keys()):
            value = metadata[key]
            if isinstance(value, np.ndarray) and value.ndim == 2:
                if value.shape[0] == self._shape[0]: # put in barcode_multiarrays
                    barcode_multiarrays[key] = metadata.pop(key)
                elif value.shape[0] == self._shape[1]: # put in feature_multiarrays
                    feature_multiarrays[key] = metadata.pop(key)
            elif isinstance(value, csr_matrix) and value.shape[0] == value.shape[1]: # is a graph
                if value.shape[0] == self._shape[0]: # put in barcode_multigraphs
                    barcode_multigraphs[key] = metadata.pop(key)
                elif value.shape[0] == self._shape[1]: # put in feature_multigraphs
                    feature_multigraphs[key] = metadata.pop(key)

        self.metadata = DataDict(metadata) # other metadata, a dictionary

        self.barcode_multiarrays = DataDict(barcode_multiarrays)
        self.feature_multiarrays = DataDict(feature_multiarrays)
        self.barcode_multigraphs = DataDict(barcode_multigraphs)
        self.feature_multigraphs = DataDict(feature_multigraphs)


    def _gen_repr_str_for_attrs(self, key: str) -> str:
        if key == "obs" or key == "obsm":
            attr_str = ""
            for attr in getattr(self, key):
                attr_type = self.get_attr_type(attr)
                attr_str += f"'{attr}', " if attr_type is None else f"'{attr}'({attr_type}), "
            return attr_str[:-2]
        else:
            return str(list(getattr(self, key).keys()))[1:-1]

    def __repr__(self) -> str:
        repr_str = f"{self.__class__.__name__} object with n_obs x n_vars = {self.barcode_metadata.shape[0]} x {self.feature_metadata.shape[0]}"
        repr_str += f"\n    Genome: {self.get_genome()}; Modality: {self.get_modality()}"
        mat_word = 'matrices' if len(self.matrices) > 1 else 'matrix'
        repr_str += f"\n    It contains {len(self.matrices)} {mat_word}: {str(list(self.matrices))[1:-1]}"
        repr_str += f"\n    It currently binds to matrix '{self._cur_matrix}' as X\n" if len(self.matrices) > 0 else "\n    It currently binds to no matrix\n"
        for key in ["obs", "var", "obsm", "varm", "obsp", "varp", "uns"]:
            fstr = self._gen_repr_str_for_attrs(key)
            if fstr != '':
                repr_str += f"\n    {key}: {fstr}"
        return repr_str


    def _update_shape(self) -> None:
        self._shape = (self.barcode_metadata.shape[0], self.feature_metadata.shape[0]) # shape


    def _is_dirty(self) -> bool:
        """ Check if any field is modified.
        """
        return self.matrices.is_dirty() or self.metadata.is_dirty() or self.barcode_multiarrays.is_dirty() or self.feature_multiarrays.is_dirty() or self.barcode_multigraphs.is_dirty() or self.feature_multigraphs.is_dirty()

    def _clear_dirty(self) -> None:
        """ Clear all dirty sets
        """
        self.matrices.clear_dirty()
        self.barcode_multiarrays.clear_dirty()
        self.feature_multiarrays.clear_dirty()
        self.barcode_multigraphs.clear_dirty()
        self.feature_multigraphs.clear_dirty()
        self.metadata.clear_dirty()


    @property
    def obs(self) -> pd.DataFrame:
        return self.barcode_metadata

    @obs.setter
    def obs(self, obs: pd.DataFrame):
        assert obs.shape[0] == 0 or obs.index.name == "barcodekey"
        self.barcode_metadata = obs
        self._update_shape()

    @property
    def obs_names(self) -> pd.Index:
        return self.barcode_metadata.index

    @obs_names.setter
    def obs_names(self, obs_names: pd.Index):
        self.barcode_metadata.index = obs_names
        self.barcode_metadata.index.name = "barcodekey"

    @property
    def var(self) -> pd.DataFrame:
        return self.feature_metadata

    @var.setter
    def var(self, var: pd.DataFrame):
        assert var.shape[0] == 0 or var.index.name == "featurekey"
        self.feature_metadata = var
        self._update_shape()

    @property
    def var_names(self) -> pd.Index:
        return self.feature_metadata.index

    @var_names.setter
    def var_names(self, var_names: pd.Index):
        self.feature_metadata.index = var_names
        self.feature_metadata.index.name = "featurekey"

    @property
    def X(self) -> Union[csr_matrix, None]:
        return self.matrices.get(self._cur_matrix, None)

    @X.setter
    def X(self, X: csr_matrix):
        self.matrices[self._cur_matrix] = X

    @property
    def obsm(self) -> Dict[str, np.ndarray]:
        return self.barcode_multiarrays

    @obsm.setter
    def obsm(self, obsm: Dict[str, np.ndarray]):
        self.barcode_multiarrays.overwrite(obsm)

    @property
    def varm(self) -> Dict[str, np.ndarray]:
        return self.feature_multiarrays

    @varm.setter
    def varm(self, varm: Dict[str, np.ndarray]):
        self.feature_multiarrays.overwrite(varm)

    @property
    def obsp(self) -> Dict[str, csr_matrix]:
        return self.barcode_multigraphs

    @obsp.setter
    def obsp(self, obsp: Dict[str, csr_matrix]):
        self.barcode_multigraphs.overwrite(obsp)

    @property
    def varp(self) -> Dict[str, csr_matrix]:
        return self.feature_multigraphs

    @varp.setter
    def varp(self, varp: Dict[str, csr_matrix]):
        self.feature_multigraphs.overwrite(varp)

    @property
    def uns(self) -> DataDict:
        return self.metadata

    @uns.setter
    def uns(self, uns: DataDict):
        self.metadata.overwrite(uns)

    @property
    def shape(self) -> Tuple[int, int]:
        return self._shape

    @shape.setter
    def shape(self, _shape: Tuple[int, int]):
        raise ValueError("Cannot set shape attribute!")


    def get_attr_type(self, attr:str) -> str:
        """ Return registered type for an attribute. If not registered, return None
        """
        if ("_attr2type" not in self.metadata) or (attr not in self.metadata["_attr2type"]):
            return None
        return self.metadata["_attr2type"][attr]

    def register_attr(self, attr: str, attr_type: str = None) -> None:
        """ Register an attribute (either in obs or obsm) with an attr_type (e.g. signature, cluster, basis, knn). Default is None (no attr_type or pop up)
        """
        if attr_type is None:
            if "_attr2type" in self.metadata:
                self.metadata["_attr2type"].pop(attr, None)
        else:
            if "_attr2type" not in self.metadata:
                self.metadata["_attr2type"] = dict()
            self.metadata["_attr2type"][attr] = attr_type


    def _polish_featurekey(self, feature_metadata: pd.DataFrame, genome: str) -> None:
        """    Remove prefixing genome strings and deduplicate feature keys
        """
        if genome != "unknown":
            # remove genome strings for 10x chemistry if > 1 genomes exist
            import re
            prefix = re.compile(f"^{genome}_+")
            if prefix.match(feature_metadata["featurekey"][0]):
                feature_metadata["featurekey"] = np.array([prefix.sub("", x) for x in feature_metadata["featurekey"].values], dtype = object)
                if "featureid" in feature_metadata:
                    feature_metadata["featureid"] = np.array([prefix.sub("", x) for x in feature_metadata["featureid"].values], dtype = object)

        if feature_metadata["featurekey"].duplicated().sum() > 0:
            # deduplicate feature keys
            keys = feature_metadata["featurekey"].values
            ids = feature_metadata["featureid"].values if "featureid" in feature_metadata else None

            from collections import Counter
            dup_ids = Counter()

            for i in range(keys.size):
                idn = dup_ids[keys[i]]
                if idn > 0:
                    # key only is duplicated
                    if ids is not None:
                        keys[i] = keys[i] + '_' + ids[i]
                        idn = dup_ids[keys[i]]
                idn += 1
                dup_ids[keys[i]] = idn
                if idn > 1:
                    keys[i] = keys[i] + f".#~{idn}" # duplicate ID starts from 2. .#~ makes it unique.


    def get_genome(self) -> str:
        """ return genome
        """
        return self.metadata.get("genome", None)


    def get_modality(self) -> str:
        """ return modality, can be either 'rna', 'atac', 'tcr', 'bcr', 'crispr', 'hashing', 'citeseq' or 'cyto' (flow cytometry / mass cytometry).
        """
        return self.metadata.get("modality", None)


    def get_uid(self) -> str:
        """ return uid used for indexing this object in a MultimodalData object. uid = genome + '-' + modality
        """
        genome = self.get_genome()
        modality = self.get_modality()
        if genome is None or modality is None:
            return None
        return genome + "-" + modality


    def list_keys(self, key_type: str = "matrix") -> List[str]:
        """ Return available keys in metadata, key_type = barcode, feature, matrix, other
        """
        if key_type == "matrix":
            return list(self.matrices)
        elif key_type == "barcode":
            return [
                self.barcode_metadata.index.name
            ] + self.barcode_metadata.columns.tolist()
        elif key_type == "feature":
            return [
                self.feature_metadata.index.name
            ] + self.feature_metadata.columns.tolist()
        elif key_type == "other":
            return list(self.metadata)
        else:
            raise ValueError(f"Unknown type {key_type}!")


    def current_matrix(self) -> str:
        """ Return current matrix as self.X
        """
        return self._cur_matrix


    def add_matrix(self, key: str, mat: csr_matrix) -> None:
        """ Add a new matrix, can be raw count or others
        """
        if key in self.matrices:
            raise ValueError(f"Matrix key '{key}' already exists!")
        if mat.shape[0] != self._shape[0]:
            raise ValueError(f"Wrong number of barcodes: matrix '{key}' has {mat.shape[0]} barcodes, which does not match with the barcode file ({self._shape[0]})!")
        if mat.shape[1] != self._shape[1]:
            raise ValueError(f"Wrong number of features: matrix '{key}' has {mat.shape[1]} features, which does not match with the feature file ({self._shape[1]})!")

        self.matrices[key] = mat


    def select_matrix(self, key: str) -> None:
        """ Select a matrix for self.X
        """
        if key not in self.matrices:
            raise ValueError(f"Matrix key '{key}' does not exist!")
        self._cur_matrix = key


    def get_matrix(self, key: str) -> csr_matrix:
        """ Return a matrix indexed by key
        """
        if key not in self.matrices:
            raise ValueError(f"Matrix key '{key}' does not exist!")
        return self.matrices[key]


    def as_float(self, key: str = None) -> None:
        """ Convert self.matrices[key] as float
        """
        key = self._cur_matrix if key is None else key
        X = self.matrices[key]
        if X.dtype == np.int32:
            X.dtype = np.float32
            orig_data = X.data.view(np.int32)
            X.data[...] = orig_data


    @run_gc
    def _inplace_subset_obs(self, index: List[bool]) -> None:
        """ Subset barcode_metadata inplace """
        if isinstance(index, pd.Series):
            index = index.values
        if index.sum() == self._shape[0]:
            return None
        self.barcode_metadata = self.barcode_metadata.loc[index].copy(deep = False)
        for key in list(self.matrices):
            self.matrices[key] = self.matrices[key][index, :]
        for key in list(self.barcode_multiarrays):
            self.barcode_multiarrays[key] = self.barcode_multiarrays[key][index]
        for key in list(self.barcode_multigraphs):
            self.barcode_multigraphs[key] = self.barcode_multigraphs[key][index][:,index]
        if "_obs_keys" in self.metadata:
            for key in self.metadata["_obs_keys"]:
                self.metadata[key] = self.metadata[key][index]
        self._update_shape()


    @run_gc
    def _inplace_subset_var(self, index: List[bool]) -> None:
        """ Subset feature_metadata inplace """
        if isinstance(index, pd.Series):
            index = index.values
        if index.sum() == self._shape[1]:
            return None
        self.feature_metadata = self.feature_metadata.loc[index].copy(deep = False)
        for key in list(self.matrices):
            self.matrices[key] = self.matrices[key][:, index]
        for key in list(self.feature_multiarrays):
            self.feature_multiarrays[key] = self.feature_multiarrays[key][index]
        for key in list(self.feature_multigraphs):
            self.feature_multigraphs[key] = self.feature_multigraphs[key][index][:,index]
        if "_var_keys" in self.metadata:
            for key in self.metadata["_var_keys"]:
                self.metadata[key] = self.metadata[key][index]
        self._update_shape()


    def separate_channels(self) -> None:
        """ Separate channel information from barcodekeys, used for 10x v2, v3 h5 and mtx as well as Optimus loom.
        """
        if self.barcode_metadata.shape[0] == 0 or self.barcode_metadata.index[0].find("-") < 0:
            return None # no data or no dash to remove

        try:
            from pegasusio.cylib.funcs import split_barcode_channel
        except ModuleNotFoundError:
            print("No module named 'pegasusio.cylib.funcs'")

        barcodes, channels = split_barcode_channel(self.barcode_metadata.index.values)

        if np.unique(channels).size > 1:
            # we have multiple channels
            self.barcode_metadata["Channel"] = channels
            barcodes = np.array([x + "-" + y for x, y in zip(channels, barcodes)], dtype = object)

        self.barcode_metadata.index = pd.Index(barcodes, name="barcodekey")



    def scan_black_list(self, black_list: Set[str]):
        """ Remove (key, value) pairs where key is in black_list
        """
        def _scan_dataframe(df: pd.DataFrame, black_list: Set[str]):
            cols = []
            for key in df.columns:
                if key in black_list:
                    cols.append(key)
            if len(cols) > 0:
                df.drop(columns = cols, inplace = True)

        def _scan_dict(mapping: dict, black_list: Set[str]):
            for key in list(mapping):
                if key in black_list:
                    del mapping[key]

        _scan_dataframe(self.barcode_metadata, black_list)
        _scan_dataframe(self.feature_metadata, black_list)

        _scan_dict(self.matrices, black_list)
        _scan_dict(self.barcode_multiarrays, black_list)
        _scan_dict(self.feature_multiarrays, black_list)
        _scan_dict(self.barcode_multigraphs, black_list)
        _scan_dict(self.feature_multigraphs, black_list)
        _scan_dict(self.metadata, black_list)


    def from_anndata(self, data: anndata.AnnData, genome: str = None, modality: str = None) -> None:
        """ Initialize from an anndata object, try best not to copy
            If genome/modality is not None, set 'genome'/'modality' as genome/modality
        """
        if data.is_view:
            data = data.copy()

        self.barcode_metadata = data.obs
        self.barcode_metadata.index.name = "barcodekey"

        self.feature_metadata = data.var
        self.feature_metadata.index.name = "featurekey"
        if "gene_ids" in self.feature_metadata:
            self.feature_metadata.rename(columns = {"gene_ids": "featureid"}, inplace = True)

        self.matrices = DataDict({"X": csr_matrix(data.X)}) # csr_matrix will not copy X.data
        if data.raw is not None:
            self.matrices["raw.X"] = csr_matrix(data.raw.X)
        for key, value in data.layers.items():
            self.matrices[key] = csr_matrix(value)

        def _create_data_dict(old_dict: dict) -> DataDict:
            from pandas.api.types import is_dict_like
            new_dict = dict()
            for key, value in old_dict.items():
                if str(type(value)).find("anndata") >= 0:
                    # This is anndata defined type
                    if is_dict_like(value):
                        new_dict[key] = dict(value)
                    else:
                        logger.warning(f"{key} is in anndata-defined data type {type(value)} and thus skipped!")
                else:
                    new_dict[key] = value
            return DataDict(new_dict)

        self.barcode_multiarrays = _create_data_dict(data.obsm)
        self.feature_multiarrays = _create_data_dict(data.varm)
        self.barcode_multigraphs = _create_data_dict(data.obsp) # Note that we do not check if data.obsp contains sparse matrix
        self.feature_multigraphs = _create_data_dict(data.varp) # Note that we do not check if data.obsp contains sparse matrix
        self.metadata = _create_data_dict(data.uns)
        _set_genome(self.metadata, genome)
        _set_modality(self.metadata, modality)


        self._cur_matrix = "X"
        self._shape = data.shape

    def to_anndata(self) -> anndata.AnnData:
        """ Convert to anndata
        """
        raw = None
        if "raw.X" in self.matrices:
            var_cols = []
            if "featureid" in self.feature_metadata:
                var_cols.append("featureid")
            raw = anndata.AnnData(X = self.matrices["raw.X"], var = self.feature_metadata[var_cols])

        layers = {}
        for key, value in self.matrices.items():
            if key != "X" and key != "raw.X":
                layers[key] = value

        return anndata.AnnData(X = self.matrices.get("X", None),
            obs = self.barcode_metadata,
            var = self.feature_metadata,
            uns = self.metadata,
            obsm = self.barcode_multiarrays,
            varm = self.feature_multiarrays,
            obsp = self.barcode_multigraphs,
            varp = self.feature_multigraphs,
            layers = layers,
            raw = raw)


    def copy(self) -> "UnimodalData":
        return UnimodalData(self.barcode_metadata.copy(),
                            self.feature_metadata.copy(),
                            deepcopy(self.matrices),
                            deepcopy(self.metadata),
                            deepcopy(self.barcode_multiarrays),
                            deepcopy(self.feature_multiarrays),
                            deepcopy(self.barcode_multigraphs),
                            deepcopy(self.feature_multigraphs),
                            self._cur_matrix)


    def __deepcopy__(self, memo: Dict):
        return self.copy()


    def _copy_view(self, viewobj: UnimodalDataView) -> "UnimodalData":
        """ In uns: Do not copy any sparse matrix or ndarrady with ndim > 1
        """
        return UnimodalData(viewobj.obs.copy(),
                            viewobj.var.copy(),
                            viewobj._copy_matrices(),
                            viewobj.uns[...],
                            viewobj.obsm[...],
                            viewobj.varm[...],
                            viewobj.obsp[...],
                            viewobj.varp[...],
                            viewobj._cur_matrix)


    def __getitem__(self, index: INDEX) -> UnimodalDataView:
        barcode_index, feature_index = _parse_index(self, index)
        return UnimodalDataView(self, barcode_index, feature_index, self._cur_matrix)


    def _update_barcode_metadata_info(
        self, row: pd.Series, attributes: Set[str], append_sample_name: bool
    ) -> None:
        """ Update barcodekey, update channel and add attributes
        """
        nsample = self.barcode_metadata.shape[0]
        sample_name = row["Sample"]

        if append_sample_name:
            barcodes = [sample_name + "-" + x for x in self.barcode_metadata.index]
            self.barcode_metadata.index = pd.Index(barcodes, name="barcodekey")
            if "Channel" in self.barcode_metadata:
                self.barcode_metadata["Channel"] = [
                    sample_name + "-" + x for x in self.barcode_metadata["Channel"]
                ]

        if "Channel" not in self.barcode_metadata:
            self.barcode_metadata["Channel"] = np.repeat(sample_name, nsample)

        for attr in attributes:
            self.barcode_metadata[attr] = np.repeat(row[attr], nsample)

        self.metadata["_sample"] = sample_name # record sample_name for merging


    def _convert_attributes_to_categorical(self, attributes: Set[str]) -> None:
        for attr in attributes:
            values = self.barcode_metadata[attr].values
            self.barcode_metadata[attr] = pd.Categorical(values, categories=natsorted(np.unique(values)))


    def _clean_tmp(self) -> dict:
        _tmp_dict = {}
        for key in ["metadata", "barcode_multiarrays", "feature_multiarrays", "barcode_multigraphs", "feature_multigraphs"]:
            obj = getattr(self, key)
            a_dict = {}
            for attr in list(obj):
                if attr.startswith("_tmp"):
                    a_dict[attr] = obj.pop(attr)
            if len(a_dict) > 0:
                _tmp_dict[key] = a_dict
        return _tmp_dict if len(_tmp_dict) > 0 else None

    def _addback_tmp(self, _tmp_dict: dict) -> None:
        for key in ["metadata", "barcode_multiarrays", "feature_multiarrays", "barcode_multigraphs", "feature_multigraphs"]:
            if key in _tmp_dict:
                getattr(self, key).update(_tmp_dict[key])



# This code is inspired by HCA zarr python codes and AnnData zarr codes

import os
import shutil
import numpy as np
import pandas as pd
from pandas.api.types import is_categorical_dtype, is_string_dtype, is_scalar, is_dict_like
from scipy.sparse import csr_matrix, issparse
import zarr
from zarr import Blosc
from natsort import natsorted
from typing import List, Dict, Tuple, Union

import logging
logger = logging.getLogger(__name__)

from pegasusio import modalities
from pegasusio import UnimodalData, VDJData, CITESeqData, CytoData, MultimodalData


CHUNKSIZE = 1000000.0
COMPRESSOR = Blosc(cname = 'lz4', clevel = 5)


def calc_chunk(shape: tuple) -> tuple:
    ndim = len(shape)
    chunks = [1] * ndim
    ords = np.argsort(shape)
    if shape[ords[0]] > 0:
        chunk_size = CHUNKSIZE
        chunk_value = -1
        for i, idx in enumerate(ords):
            if chunk_value < 0:
                if shape[idx] ** (ndim - i) < chunk_size:
                    chunks[idx] = shape[idx]
                    chunk_size /= chunks[idx]
                else:
                    chunk_value = int(np.ceil(chunk_size ** (1.0 / (ndim - i))))
                    chunks[idx] = chunk_value
            else:
                chunks[idx] = chunk_value
    return tuple(chunks)


def check_and_remove_existing_path(path: str) -> None:
    if os.path.exists(path):
        if os.path.isdir(path):
            shutil.rmtree(path)
            logger.warning(f"Detected and removed pre-existing directory {path}.")
        else:
            assert os.path.isfile(path)
            os.unlink(path)
            logger.warning(f"Detected and removed pre-existing file {path}.")


class ZarrFile:
    def __init__(self, path: str, mode: str = 'r', storage_type: str = None) -> None:
        """ Initialize a Zarr file, if mode == 'w', create an empty one, otherwise, load from path
        path : `str`, path for the zarr object.
        storage_type : `str`, currently only support 'ZipStore' and 'NestedDirectoryStore'. If None, use 'NestedDirectoryStore' by default.
        """
        self.store = self.root = None

        if storage_type is None:
            storage_type = 'NestedDirectoryStore'

        if mode == 'w':
            # Create a new zarr file
            check_and_remove_existing_path(path)
            self.store = zarr.ZipStore(path, mode = 'w') if storage_type == 'ZipStore' else zarr.NestedDirectoryStore(path)
            self.root = zarr.group(self.store, overwrite = True)
        else:
            # Load existing zarr file
            self.store = zarr.NestedDirectoryStore(path) if os.path.isdir(path) else zarr.ZipStore(path, mode = 'r')
            if mode == 'a' and isinstance(self.store, zarr.ZipStore):
                self._to_directory()
            self.root = zarr.open(self.store, mode = mode)


    def __del__(self):
        if self.store is not None and hasattr(self.store, 'close'):
            self.store.close()


    def _to_zip(self):
        if not isinstance(self.store, zarr.ZipStore):
            zip_path = self.store.path + '.zip'
            zip_store = zarr.ZipStore(zip_path, mode = 'w')
            zarr.copy_store(self.store, zip_store)
            zip_store.close()

            shutil.rmtree(self.store.path)

            self.store = zarr.ZipStore(zip_path, mode = 'r')
            self.root = zarr.open_group(self.store, mode = 'r')


    def _to_directory(self):
        orig_path = self.store.path

        if not orig_path.endswith('.zip'):
            self.store.close()
            zip_path = orig_path + '.zip'
            check_and_remove_existing_path(zip_path)
            os.replace(orig_path, zip_path)
            self.store = zarr.ZipStore(zip_path, mode = 'r')
        else:
            zip_path = orig_path

        dest_path = zip_path[:-4]
        check_and_remove_existing_path(dest_path)
        dir_store = zarr.NestedDirectoryStore(dest_path)
        zarr.copy_store(self.store, dir_store)
        self.store.close()
        os.remove(zip_path)

        self.store = dir_store
        self.root = zarr.open_group(self.store)

        logger.info(f"Converted ZipStore zarr file {orig_path} to NestedDirectoryStore {dest_path}.")


    def read_csr(self, group: zarr.Group) -> csr_matrix:
        return csr_matrix((group['data'][...], group['indices'][...], group['indptr'][...]), shape = group.attrs['shape'])


    def read_series(self, group: zarr.Group, name: str) -> Union[pd.Categorical, np.ndarray]:
        if 'ordered' in group[name].attrs:
            # categorical column
            return pd.Categorical.from_codes(group[name][...], categories = group[f'_categories/{name}'][...], ordered = group[name].attrs['ordered'])
        else:
            return group[name][...]


    def read_dataframe(self, group: zarr.Group) -> pd.DataFrame:
        columns = group.attrs.get('columns', None)
        if columns is None:
            columns = [col for col in group.array_keys() if col != '_index']

        data = {col: self.read_series(group, col) for col in columns}
        _index = self.read_series(group, '_index')
        index = pd.Index(_index, name = group.attrs['index_name'], dtype = _index.dtype)
        df = pd.DataFrame(data = data, index = index) # if add columns = columns, the generation will be slow
        return df


    def read_array(self, group: zarr.Group, name: str) -> np.ndarray:
        return group[name][...]


    def read_record_array(self, group: zarr.Group) -> np.recarray:
        columns = group.attrs.get('columns', None)
        if columns is None:
            columns = [col for col in group.array_keys()]

        array = np.rec.fromarrays([group[col][...] for col in columns], names = columns)
        return array


    def read_composite_list(self, group: zarr.Group) -> list:
        assert '_size' in group.attrs
        res_list = [None] * group.attrs['_size']

        if 'scalar' in group.attrs:
            scalar_dict = group.attrs['scalar']
            for i, value in scalar_dict.items():
                res_list[int(i)] = value

        for key in group. array_keys():
            res_list[int(key)] = self.read_array(group, key)

        for key in group.group_keys():
            sub_group = group[key]
            data_type = sub_group.attrs['data_type']
            value = None
            if data_type == 'data_frame':
                value = self.read_dataframe(sub_group)
            elif data_type == 'record_array':
                value = self.read_record_array(sub_group)
            elif data_type == 'csr_matrix':
                value = self.read_csr(sub_group)
            elif data_type == 'dict':
                value = self.read_mapping(sub_group)
            else:
                assert data_type == 'composite_list'
                value = self.read_composite_list(sub_group)
            res_list[int(key)] = value

        return res_list


    def read_mapping(self, group: zarr.Group) -> dict:
        res_dict = {}

        if 'scalar' in group.attrs:
            res_dict.update(group.attrs['scalar'])
        
        for key in group.array_keys():
            res_dict[key] = self.read_array(group, key)
        
        for key in group.group_keys():
            sub_group = group[key]
            data_type = sub_group.attrs['data_type']
            value = None
            if data_type == 'data_frame':
                value = self.read_dataframe(sub_group)
            elif data_type == 'record_array':
                value = self.read_record_array(sub_group)
            elif data_type == 'csr_matrix':
                value = self.read_csr(sub_group)
            elif data_type == 'dict':
                value = self.read_mapping(sub_group)
            else:
                assert data_type == 'composite_list'
                value = self.read_composite_list(sub_group)
            res_dict[key] = value

        return res_dict


    def read_unimodal_data(self, group: zarr.Group) -> UnimodalData:
        """ Read UnimodalData
        """
        metadata = self.read_mapping(group['metadata'])
        assert 'genome' in metadata
        if 'modality' not in metadata:
            assert metadata.get('experiment_type', 'none') in modalities
            metadata['modality'] = metadata.pop('experiment_type')

        DataClass = UnimodalData
        modality = metadata['modality']
        if modality == 'tcr' or modality == 'bcr':
            DataClass = VDJData
        elif modality == 'citeseq':
            DataClass = CITESeqData
        elif modality == 'cyto':
            DataClass = CytoData

        unidata = DataClass(barcode_metadata = self.read_dataframe(group['barcode_metadata']),
                            feature_metadata = self.read_dataframe(group['feature_metadata']),
                            matrices = self.read_mapping(group['matrices']),
                            metadata = metadata,
                            barcode_multiarrays = self.read_mapping(group['barcode_multiarrays']),
                            feature_multiarrays = self.read_mapping(group['feature_multiarrays']),
                            barcode_multigraphs = self.read_mapping(group['barcode_multigraphs']) if 'barcode_multigraphs' in group else dict(), # for backward-compatibility
                            feature_multigraphs = self.read_mapping(group['feature_multigraphs']) if 'feature_multigraphs' in group else dict(), # for backward-compatibility
                            )

        if group.attrs.get('_cur_matrix', None) is not None:
            unidata.select_matrix(group.attrs['_cur_matrix'])

        return unidata


    def read_multimodal_data(self, attach_zarrobj = False) -> MultimodalData:
        """ Read MultimodalData
        """
        data = MultimodalData()
        for key, group in self.root.groups():
            unidata = self.read_unimodal_data(group)
            data.add_data(unidata)

        if self.root.attrs.get('_selected', None) is not None:
            data.select_data(self.root.attrs['_selected'])

        if attach_zarrobj:
            data._zarrobj = self

        return data


    def write_csr(self, parent: zarr.Group, name: str, matrix: csr_matrix) -> None:
        group = parent.create_group(name, overwrite = True)
        group.attrs.update(data_type = 'csr_matrix', shape = matrix.shape)
        group.create_dataset('data', data = matrix.data, shape = matrix.data.shape, chunks = calc_chunk(matrix.data.shape), dtype = matrix.data.dtype, compressor = COMPRESSOR, overwrite = True)
        group.create_dataset('indices', data = matrix.indices, shape = matrix.indices.shape, chunks = calc_chunk(matrix.indices.shape), dtype = matrix.indices.dtype, compressor = COMPRESSOR, overwrite = True)
        group.create_dataset('indptr', data = matrix.indptr, shape = matrix.indptr.shape, chunks = calc_chunk(matrix.indptr.shape), dtype = matrix.indptr.dtype, compressor = COMPRESSOR, overwrite = True)


    def write_series(self, group: zarr.Group, name: str, array: np.ndarray) -> None:
        if not is_categorical_dtype(array) and name != '_index' and is_string_dtype(array):
            keywords = set(array)
            if len(keywords) <= array.size / 10.0: # at least 10x reduction
                array = pd.Categorical(array, categories = natsorted(keywords))

        if is_categorical_dtype(array):
            # write category keys
            categories = group.require_group('_categories', overwrite = False)
            values = array.categories.values
            if isinstance(values[0], bytes):
                values = np.array([x.decode() for x in values], dtype = object)
            self.write_array(categories, name, values)
            # write codes
            codes_arr = group.create_dataset(name, data = array.codes, shape = array.codes.shape, chunks = calc_chunk(array.codes.shape), dtype = array.codes.dtype, compressor = COMPRESSOR, overwrite = True)
            codes_arr.attrs['ordered'] = array.ordered
        else:
            self.write_array(group, name, array)


    def write_dataframe(self, parent: zarr.Group, name: str, df: pd.DataFrame) -> None:
        group = parent.create_group(name, overwrite = True)
        attrs_dict = {'data_type' : 'data_frame', 'columns' : list(df.columns)}
        attrs_dict['index_name'] = df.index.name if df.index.name is not None else 'index'
        self.write_series(group, '_index', df.index.values)
        for col in df.columns:
            self.write_series(group, col, df[col].values)
        group.attrs.update(**attrs_dict)


    def write_array(self, group: zarr.Group, name: str, array: np.ndarray) -> None:
        dtype = str if array.dtype.kind == 'O' else array.dtype
        group.create_dataset(name, data = array, shape = array.shape, chunks = calc_chunk(array.shape), dtype = dtype, compressor = COMPRESSOR, overwrite = True)


    def write_record_array(self, parent: zarr.Group, name: str, array: np.recarray) -> None:
        group = parent.create_group(name, overwrite = True) 
        attrs_dict = {'data_type' : 'record_array', 'columns' : list(array.dtype.names)}
        for col in array.dtype.names:
            self.write_array(group, col, array[col])
        group.attrs.update(**attrs_dict)


    def write_composite_list(self, parent: zarr.Group, name: str, values: Union[list, tuple]) -> None:
        group = parent.create_group(name, overwrite = True)
        scalar_dict = {}

        for i, value in enumerate(values):
            if is_scalar(value):
                if type(value).__module__ == 'numpy':
                    value = value.item()
                scalar_dict[str(i)] = value
            elif isinstance(value, np.ndarray):
                if value.dtype.kind == 'V':
                    self.write_record_array(group, str(i), value)
                else:
                    self.write_array(group, str(i), value)
            elif isinstance(value, pd.DataFrame):
                self.write_dataframe(group, str(i), value)
            elif is_dict_like(value):
                self.write_mapping(group, str(i), value)
            elif issparse(value):
                assert isinstance(value, csr_matrix)
                self.write_csr(group, str(i), value)
            else:
                # assume value is either list or tuple
                assert type(value) in {list, tuple}
                is_comp_list = sum([is_scalar(x) for x in value]) < len(value)
                if is_comp_list:
                    self.write_composite_list(group, str(i), value)
                else:
                    # converting it to np.ndarray
                    self.write_array(group, str(i), value.astype(str) if is_categorical_dtype(value) else np.array(value))

        attrs_dict = {'data_type': 'composite_list', '_size': len(values)}
        if len(scalar_dict) > 0:
            attrs_dict['scalar'] = scalar_dict
        group.attrs.update(**attrs_dict)


    def write_mapping(self, parent: zarr.Group, name: str, mapping: dict, overwrite = True) -> None:
        group = None
        if overwrite:
            group = parent.create_group(name, overwrite = True)
        else:
            group = parent.require_group(name, overwrite = False) # throw an error if name in array_keys()

        scalar_dict = group.attrs.pop('scalar', {})

        def _write_one_pair(key, value):
            if is_scalar(value):
                if type(value).__module__ == 'numpy':
                    value = value.item()
                scalar_dict[key] = value
            elif isinstance(value, np.ndarray):
                if value.dtype.kind == 'V':
                    self.write_record_array(group, key, value)
                else:
                    self.write_array(group, key, value)
            elif isinstance(value, pd.DataFrame):
                self.write_dataframe(group, key, value)
            elif is_dict_like(value):
                self.write_mapping(group, key, value)
            elif issparse(value):
                assert isinstance(value, csr_matrix)
                self.write_csr(group, key, value)
            else:
                # assume value is either list or tuple
                assert type(value) in {list, tuple}
                is_comp_list = sum([is_scalar(x) for x in value]) < len(value)
                if is_comp_list:
                    self.write_composite_list(group, key, value)
                else:
                    # converting it to np.ndarray
                    self.write_array(group, key, value.astype(str) if is_categorical_dtype(value) else np.array(value))

        if overwrite:
            for key, value in mapping.items():
                _write_one_pair(key, value)
        else:
            for key in mapping.deleted:
                if key in scalar_dict:
                    del scalar_dict[key]
                else:
                    del group[key]
            for key in mapping.modified:
                _write_one_pair(key, mapping[key])

        attrs_dict = {'data_type' : 'dict'}
        if len(scalar_dict) > 0:
            attrs_dict['scalar'] = scalar_dict
        group.attrs.update(**attrs_dict)


    def write_unimodal_data(self, parent: zarr.Group, name: str, data: UnimodalData, overwrite: bool = True) -> None:
        """ Write UnimodalData
            overwrite means if overwrite the whole unimodal data; meaning is different from require_group overwrite --- if the name in array_keys(), take it and make it as a group
        """
        group = None
        if overwrite:
            group = parent.create_group(name, overwrite = True)
        else:
            group = parent.require_group(name, overwrite = False) # throw an error if name in array_keys()

        attrs_dict = {'data_type': 'UnimodalData', '_cur_matrix': data.current_matrix()}
        group.attrs.update(**attrs_dict)

        self.write_dataframe(group, 'barcode_metadata', data.barcode_metadata)
        self.write_dataframe(group, 'feature_metadata', data.feature_metadata)

        if overwrite or data.matrices.is_dirty():
            self.write_mapping(group, 'matrices', data.matrices, overwrite = overwrite)
        if overwrite or data.metadata.is_dirty():
            self.write_mapping(group, 'metadata', data.metadata, overwrite = overwrite)
        if overwrite or data.barcode_multiarrays.is_dirty():
            self.write_mapping(group, 'barcode_multiarrays', data.barcode_multiarrays, overwrite = overwrite)
        if overwrite or data.feature_multiarrays.is_dirty():
            self.write_mapping(group, 'feature_multiarrays', data.feature_multiarrays, overwrite = overwrite)
        if overwrite or data.barcode_multigraphs.is_dirty():
            self.write_mapping(group, 'barcode_multigraphs', data.barcode_multigraphs, overwrite = overwrite)
        if overwrite or data.feature_multigraphs.is_dirty():
            self.write_mapping(group, 'feature_multigraphs', data.feature_multigraphs, overwrite = overwrite)


    def write_multimodal_data(self, data: MultimodalData, overwrite: bool = True) -> None:
        """ Write MultimodalData
        """
        if overwrite:
            for key in data.list_data():
                self.write_unimodal_data(self.root, key, data.get_data(key), overwrite = True)
        else:
            for key in data.data.deleted:
                del self.root[key]
            for key in data.data.accessed:
                self.write_unimodal_data(self.root, key, data.get_data(key), overwrite = key in data.data.modified)            
        self.root.attrs['_selected'] = data._selected



"""
PegasusIO is the IO component of Pegasus and a multi-modality single-cell genomics IO package. It is a command line tool, a python package and a base for Cloud-based analysis workflows.

Usage:
  pegasusio <command> [<args>...]
  pegasusio -h | --help
  pegasusio -v | --version

Sub-commands:
  aggregate_matrix        Aggregate multi-modality single-cell genomics data into one data object. It also enables users to import metadata into the aggregated data.

Options:
  -h, --help          Show help information.
  -v, --version       Show version.

Description:
  This is PegasusIO, the IO component of Pegasus.
"""

from docopt import docopt
from docopt import DocoptExit
from . import __version__ as VERSION
from . import commands


def main():
    args = docopt(__doc__, version=VERSION, options_first=True)

    command_name = args["<command>"]
    command_args = args["<args>"]
    command_args.insert(0, command_name)

    try:
        command_class = getattr(commands, command_name)
    except AttributeError:
        print("Unknown command {cmd}!".format(cmd=command_name))
        raise DocoptExit()

    command = command_class(command_args)
    command.execute()


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd
import itertools
from scipy.sparse import csr_matrix

from pegasusio import VDJData, MultimodalData


def load_10x_vdj_file(input_csv: str, genome: str = None, modality: str = None) -> MultimodalData:
    """Load VDJ data from a 10x CSV file

    Parameters
    ----------

    input_csv : `str`
        The CSV file, gzipped or not, containing the count matrix.
    genome : `str`, optional (default None)
        The genome reference. If None, use "unknown" instead.
    modality: `str`, optional (default None)
        Modality. It should be automatically detected from the CSV file. If not None and the detected modality is not the same as the one users' provide, report an error.

    Returns
    -------

    A MultimodalData object containing a (genome, VDJData) pair.

    Examples
    --------
    >>> io.load_csv_file('vdj_t_all_contig_annotations.csv', genome = 'GRCh38_tcr')
    """
    try:
        from pegasusio.cylib.funcs import convert_10x_vdj_to_vdjdata
    except ModuleNotFoundError:
        print("No module named 'pegasusio.cylib.funcs'")
        
    df = pd.read_csv(input_csv, na_filter = False) # Otherwise, '' will be converted to NaN
    idx = df["productive"] == (True if df["productive"].dtype.kind == "b" else "True")
    df = df[idx]
    df.sort_values(by = ["barcode", "umis"], ascending = [True, False], inplace = True, kind = "mergesort") # sort barcode and make sure it is stable

    feature_name = [x for x in df["chain"].value_counts().index if x != "Multi"][0]
    modal = None
    if feature_name in VDJData._features["tcr"]:
        modal = "tcr"
    elif feature_name in VDJData._features["bcr"]:
        modal = "bcr"
    else:
        raise ValueError(f"Unknown feature '{feature_name}' detected!")

    if (modality is not None) and (modality != modal):
        raise ValueError(f"Detected modality '{modal}' does not match user-provided modality '{modality}'!")
    modality = modal

    # Set up feature keys
    feature_metadata = {"featurekey": [x + (str(y + 1) if y > 0 else "") for x, y in itertools.product(VDJData._features[modality], range(VDJData._n_contigs))]}
    fid2pos = {}
    for i, value in enumerate(feature_metadata["featurekey"]):
        fid2pos[value] = i

    n_barcodes = df["barcode"].nunique()

    barcodes, is_cell, mats, strarrs = convert_10x_vdj_to_vdjdata(df["barcode"].values,
                                                                  df[VDJData._matrix_keywords[0:4] + ["is_cell"]].values.astype(np.int32),
                                                                  df[VDJData._matrix_keywords[4:] + ["chain"]].values,
                                                                  fid2pos, n_barcodes, VDJData._n_contigs)

    barcode_metadata = {"barcodekey": barcodes, "is_cell": is_cell}

    matrices = {}
    for i, keyword in enumerate(VDJData._matrix_keywords):
        mat = mats[i]
        if keyword == "high_confidence":
            mat = mat.astype(np.bool_)
        matrices[keyword] = csr_matrix(mat)

    genome = "unknown" if genome is None else genome
    metadata = {"genome": genome, "modality": modality}
    for i, keyword in enumerate(VDJData._uns_keywords):
        metadata[keyword] = strarrs[i]

    vdjdata = VDJData(barcode_metadata, feature_metadata, matrices, metadata)
    vdjdata.separate_channels()
    data = MultimodalData(vdjdata)

    return data



from setuptools import setup, find_packages
from setuptools.extension import Extension
from Cython.Build import cythonize
from codecs import open
from pathlib import Path
import os


here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

extensions = [
    Extension("pegasusio.cylib.funcs", ["ext_modules/fast_funcs.pyx"]),
    Extension("pegasusio.cylib.io", ["ext_modules/io_funcs.pyx"])
]

setup(
    name="pegasusio",
    use_scm_version=True,
    zip_safe=False,
    description="Pegasusio is a Python package for reading / writing single-cell genomics data",
    long_description=long_description,
    url="https://github.com/klarman-cell-observatory/pegasusio",
    author="Yiming Yang, Joshua Gould and Bo Li",
    author_email="cumulus-support@googlegroups.com",
    classifiers=[  # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 4 - Beta",
        "Framework :: Jupyter",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="single cell/nucleus genomics data reading and writing",
    packages=find_packages(),
    ext_modules=cythonize(extensions),
    setup_requires=["Cython", "setuptools_scm"],
    install_requires=[
        l.strip() for l in Path("requirements.txt").read_text("utf-8").splitlines()
    ],
    python_requires="~=3.6",
    entry_points={"console_scripts": ["pegasusio=pegasusio.__main__:main"]},
)



from collections.abc import MutableMapping



class DataDict(MutableMapping):
    def __init__(self, initial_data: dict = None):
        self.mapping = initial_data if initial_data is not None else dict()

        self.init_keys = set(self.mapping) # record keys when data is loaded
        self.modified = set() # record keys that are modified
        self.deleted = None # will record keys need to be removed from file, only calculated in is_dirty
        self._tmp = None # temporary value to cache new init_keys

        self._overwrite = False # if overwrite happened

    def __getitem__(self, key):
        return self.mapping[key]

    def __setitem__(self, key, value):
        self.mapping[key] = value
        self.modified.add(key)

    def __delitem__(self, key):
        del self.mapping[key]
        self.modified.discard(key)

    def __iter__(self):
        return iter(self.mapping)

    def __len__(self):
        return len(self.mapping)

    def __repr__(self):
        return f"DataDict object with keys: {str(list(self.mapping))[1:-1]}"

    def is_dirty(self):
        """ DataDict should not be updated after calling is_dirty and before calling clear_dirty """
        if self.deleted is None:
            self._tmp = set(self.mapping)
            self.deleted = self.init_keys - self._tmp
        return self._overwrite or len(self.deleted) > 0 or len(self.modified) > 0

    def clear_dirty(self):
        self.init_keys = self._tmp
        self.modified.clear()
        self.deleted = self._tmp = None
        self._overwrite = False

    def overwrite(self, mapping: dict):
        """ Completely overwrite the DataDict """
        self._overwrite = True
        self.mapping = mapping
        self.init_keys.clear()
        self.modified.clear()
        self.deleted = self._tmp = None



class MultiDataDict(MutableMapping):
    def __init__(self):
        self.mapping = dict()

        self.init_keys = set() # record keys when data is loaded
        self.accessed = set() # record keywords for accessed UnimodalData
        self.modified = set() # record keys that are modified
        self.deleted = self._tmp = None

    def __getitem__(self, key):
        value = self.mapping[key]
        self.accessed.add(key)
        return value

    def __setitem__(self, key, value):
        self.mapping[key] = value
        self.accessed.add(key)
        self.modified.add(key)

    def __delitem__(self, key):
        del self.mapping[key]
        self.accessed.discard(key)
        self.modified.discard(key)

    def __iter__(self):
        return iter(self.mapping)

    def __len__(self):
        return len(self.mapping)

    def __repr__(self):
        return f"MultiDataDict object with keys: {str(list(self.mapping))[1:-1]}"

    def kick_start(self, data_key):
        """ Since UnimodalData are added sequentially when loading the object, kick_start marks the finish of loading: we begin to track changes now """
        self.init_keys = set(self.mapping)
        self.accessed.clear()
        self.accessed.add(data_key)
        self.modified.clear()

    def is_dirty(self):
        self._tmp = set(self.mapping)
        self.deleted = self.init_keys - self._tmp

        if len(self.deleted) > 0 or len(self.modified) > 0:
            return True

        for key in self.accessed:
            if self.mapping[key]._is_dirty():
                return True

        return False

    def clear_dirty(self, data_key):
        for key in self.accessed:
            self.mapping[key]._clear_dirty()
        self.init_keys = self._tmp
        self.accessed.clear()
        self.accessed.add(data_key)
        self.modified.clear()
        self.deleted = self._tmp = None



import time
import gc
import functools


def timer(logger = None, precision: int = 2):
	""" Log the time spent for the function. """
	def inner_decorator(func):
		@functools.wraps(func)
		def wrapper_timer(*args, **kwargs):
			start = time.perf_counter()
			result = func(*args, **kwargs)
			end = time.perf_counter()
			message = f"Function '{func.__name__}' finished in {{:.{precision}f}}s.".format(end - start)

			if logger is None:
				print(message)
			else:
				logger.info(message)

			return result
		return wrapper_timer
	return inner_decorator


def run_gc(func):
	""" Run garbage collector """
	@functools.wraps(func)
	def wrapper_run_gc(*args, **kwargs):
		result = func(*args, **kwargs)
		gc.collect()
		return result
	return wrapper_run_gc
